# Copyright (C) 2021,2022,2023,2024,2025,2026 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Shared argument parsing for xsget and xstxt."""

import argparse
import platform
import sys
from collections.abc import Sequence
from typing import Any

from xsget import __version__


class EnvironmentAction(argparse.Action):
    """Show environment details action."""

    def __init__(self, **kwargs: Any) -> None:  # noqa: ANN401
        """Initialize the EnvironmentAction."""
        kwargs["nargs"] = 0
        super().__init__(**kwargs)

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        _namespace: argparse.Namespace,
        _values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        """Print environment details and exit.

        Args:
            parser: The argument parser.
            namespace: The namespace to store the parsed arguments.
            values: The argument values.
            option_string: The option string that triggered this action.
        """
        sys_version = sys.version.replace("\n", "")
        env = (
            f"xsget: {__version__}\n"
            f"python: {sys_version}\n"
            f"platform: {platform.platform()}\n"
        )
        parser._print_message(env, sys.stdout)
        parser.exit()


def create_base_parser(
    prog: str,
    description: str,
    epilog: str,
) -> argparse.ArgumentParser:
    """Create a base parser with common arguments."""
    parser = argparse.ArgumentParser(
        prog=prog,
        add_help=False,
        description=description,
        epilog=epilog,
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
            prog,
            max_help_position=8,
        ),
    )

    parser.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        dest="quiet",
        help="suppress all logging",
    )

    parser.add_argument(
        "-e",
        "--env",
        default=False,
        action=EnvironmentAction,
        dest="env",
        help="print environment information for bug reporting",
    )

    parser.add_argument(
        "-d",
        "--debug",
        default=False,
        action="store_true",
        dest="debug",
        help="show debugging log and stacktrace",
    )

    parser.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="show this help message and exit",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    return parser



import argparse
from unittest.mock import patch

import pytest

from xsget import config


def create_config_file(path, content):
    with open(path, "w") as f:
        f.write(content)


def test_load_config_success(tmp_path):
    config_path = tmp_path / "test_config.toml"
    content = 'config_version = 3\nurl = "http://example.com"'
    create_config_file(config_path, content)

    args = argparse.Namespace(config=str(config_path), generate_config=None)

    dummy_template = 'config_version = 3\nurl = "default"'

    with patch("xsget.config.read_text", return_value=dummy_template):
        result = config.load_or_create_config(args, "xsget")

    assert result["url"] == "http://example.com"
    assert result["config_version"] == 3


def test_load_config_corrupted(tmp_path):
    config_path = tmp_path / "empty_config.toml"
    create_config_file(config_path, "")

    args = argparse.Namespace(config=str(config_path), generate_config=None)

    with pytest.raises(config.ConfigFileCorruptedError):
        config.load_or_create_config(args, "xsget")


def test_create_config_exists(tmp_path):
    config_path = tmp_path / "existing.toml"
    create_config_file(config_path, "content")

    args = argparse.Namespace(config=None, generate_config=str(config_path))

    with pytest.raises(config.ConfigFileExistsError):
        config.load_or_create_config(args, "xsget")


def test_upgrade_config(tmp_path):
    config_path = tmp_path / "old_config.toml"
    # Old version
    create_config_file(
        config_path, 'config_version = 1\nurl = "http://old.com"'
    )

    args = argparse.Namespace(config=str(config_path), generate_config=None)

    # Template has version 3
    dummy_template = 'config_version = 3\nurl = "default"'

    with patch("xsget.config.read_text", return_value=dummy_template):
        # We need to mock datetime to have deterministic backup filename
        with patch("xsget.config.dt") as mock_dt:
            mock_dt.now.return_value.strftime.return_value = "20000101_120000"

            result = config.load_or_create_config(args, "xsget")

    # Check if backup was created
    backup_path = tmp_path / "old_config_20000101_120000_backup.toml"
    assert backup_path.exists()
    assert "config_version = 1" in backup_path.read_text()

    # Check if original file was updated
    assert config_path.exists()
    new_content = config_path.read_text()
    assert "config_version = 3" in new_content
    # The upgrade process loads old config into Namespace, calls _create_config
    # _create_config loads template, overrides with values from Namespace.
    # So 'url' should be preserved if it was in config_dict.
    # So 'url' should be preserved if it was in config_dict.
    assert 'url = "http://old.com"' in new_content


def test_generating_default_config_file(script_runner):
    default_url = "http://localhost"
    ret = script_runner("xsget", default_url, "-g")
    assert "Create config file: xsget.toml" in ret.stdout
    assert (
        "Cannot connect to host localhost:80 "
        "ssl:default [Connect call failed ('127.0.0.1', 80)]" in ret.stdout
    )



# Copyright (C) 2021,2022,2023,2024,2025,2026 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114,C0116

import argparse
import asyncio
import logging
from pathlib import Path
from unittest.mock import ANY, AsyncMock, MagicMock, patch

import aiohttp
import pytest
from aioresponses import aioresponses
from playwright.async_api import Error as PlaywrightError
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

from xsget.xsget import (
    _main_async_logic,
    build_parser,
    extract_urls,
    fetch_url_by_aiohttp,
    fetch_url_by_browser,
    fetch_urls,
    filter_urls,
    http_headers,
    main,
    run,
    url_to_filename,
)

DEFAULT_URL = "http://localhost"


def test_url_to_filename():
    expected = [
        ("http://a.com", "index.html"),
        ("http://a.com/", "index.html"),
        ("http://a.com/123", "123.html"),
        ("http://a.com/123/456", "456.html"),
        ("http://a.com/123/456/789", "789.html"),
        ("http://a.com/123.html", "123.html"),
        ("http://a.com/123.html?abc=def", "123.html"),
    ]
    for url, filename in expected:
        assert url_to_filename(url) == filename

    expected = [
        ("http://a.com/123?id=aaa", "id", "aaa.html"),
        ("http://a.com/456.php?tid=abc", "tid", "abc.html"),
        ("http://a.com/789.php?test=xyz&id=aaa", "id", "aaa.html"),
    ]
    for url, url_param, filename in expected:
        assert url_to_filename(url, url_param) == filename


def test_extract_urls():
    html = """
        <html>
        <body>
        <div class="toc">
            <a href="http://a.com/123"/>a</a>
            <a href="http://a.com/123/789.html"/>b</a>
            <a href="//a.com/987"/>c</a>
            <a href="/123/456"/>d</a>
            <a href="/123/654.html"/>e</a>
        </div>
        </body>
        </html>
    """

    expected_urls = [
        "http://a.com/123",
        "http://a.com/123/789.html",
        "http://a.com/987",
        "http://a.com/123/456",
        "http://a.com/123/654.html",
    ]

    css_paths = [
        "html body div.toc a",
        "html body div a",
        "body div.toc a",
        "div.toc a",
        "div a",
        "a",
    ]
    for css_path in css_paths:
        config = argparse.Namespace(
            url="http://a.com/123",
            link_css_path=css_path,
        )
        assert extract_urls(html, config) == expected_urls


def test_user_agent():
    user_agent = http_headers()["User-Agent"]
    assert "Mozilla/5.0" in user_agent


@patch("sys.stdin")
def test_build_parser_url_arg(mock_stdin):
    mock_stdin.isatty.return_value = True
    parser = build_parser([])
    with pytest.raises(SystemExit):
        parser.parse_args([])

    mock_stdin.isatty.return_value = False
    mock_stdin.read.return_value = "http://piped.url\n"
    parser = build_parser([])
    args = parser.parse_args([])
    assert args.url == "http://piped.url"

    mock_stdin.isatty.return_value = False
    mock_stdin.read.return_value = "http://piped.url\n"
    parser = build_parser(["http://cli.url"])
    args = parser.parse_args(["http://cli.url"])
    assert args.url == "http://cli.url"

    mock_stdin.isatty.return_value = True
    parser = build_parser(["-c"])
    args = parser.parse_args(["-c"])
    assert args.url == ""
    assert args.config == "xsget.toml"

    parser = build_parser(["-e"])
    url_action = next((a for a in parser._actions if a.dest == "url"), None)  # noqa: SLF001
    assert url_action is None


async def test_fetch_url_by_aiohttp(tmp_path, caplog):
    session = aiohttp.ClientSession()
    test_url = "http://test.com/page1"
    test_content = "<html>test content</html>"
    output_dir = tmp_path / "output"
    config = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(output_dir),
    )

    with aioresponses() as mocked:
        mocked.get(
            test_url,
            status=200,
            body=test_content,
            headers={"Content-Type": "text/html; charset=utf-8"},
        )

        with caplog.at_level(logging.INFO):
            result = await fetch_url_by_aiohttp(session, test_url, config)
            assert result is None

        expected_file = output_dir / "page1.html"
        assert expected_file.exists()
        assert expected_file.read_text(encoding="utf-8") == test_content

        assert "Fetching http://test.com/page1" in caplog.text
        assert f"-> Saving {expected_file}" in caplog.text

    with aioresponses() as mocked:
        mocked.get(test_url, status=404)

        with caplog.at_level(logging.ERROR):
            result = await fetch_url_by_aiohttp(session, test_url, config)
            assert result is None
            assert (
                "error: 404, message='Not Found', url='http://test.com/page1'"
                in caplog.text
            )

    invalid_url = "invalid url"
    with pytest.raises(RuntimeError, match="invalid url: invalid%20url"):
        await fetch_url_by_aiohttp(session, invalid_url, config)
    await session.close()


async def test_fetch_url_by_browser_success(tmp_path, caplog):
    test_url = "http://test.com/page1"
    test_content = "<html>browser content</html>"
    output_dir = tmp_path / "output"
    config = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(output_dir),
        browser_delay=100,
    )

    mock_response = AsyncMock()
    mock_response.header_value.return_value = "text/html; charset=iso-8859-1"
    mock_page = AsyncMock()
    mock_page.goto.return_value = mock_response
    mock_page.content.return_value = test_content
    mock_context = AsyncMock()
    mock_context.new_page.return_value = mock_page

    with caplog.at_level(logging.INFO):
        await fetch_url_by_browser(mock_context, test_url, config)

    mock_context.new_page.assert_called_once()
    mock_page.wait_for_timeout.assert_called_once_with(config.browser_delay)
    mock_page.goto.assert_called_once_with(test_url)
    mock_page.content.assert_called_once()
    mock_page.close.assert_called_once()

    expected_file = output_dir / "page1.html"
    assert expected_file.exists()
    assert expected_file.read_text(encoding="iso-8859-1") == test_content
    assert f"Fetch: {test_url} -> save: {expected_file}" in caplog.text


async def test_fetch_url_by_browser_encoding_fallback(tmp_path, caplog):
    test_url = "http://test.com/page2"
    test_content = "<html>browser content</html>"
    output_dir = tmp_path / "output_fallback"
    config = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(output_dir),
        browser_delay=0,
    )

    mock_response = AsyncMock()
    mock_response.header_value.return_value = "a=b;c"
    mock_page = AsyncMock()
    mock_page.goto.return_value = mock_response
    mock_page.content.return_value = test_content
    mock_context = AsyncMock()
    mock_context.new_page.return_value = mock_page
    with caplog.at_level(logging.WARNING):
        await fetch_url_by_browser(mock_context, test_url, config)

    expected_file = output_dir / "page2.html"
    assert expected_file.exists()
    assert expected_file.read_text(encoding="utf-8") == test_content
    assert "Could not parse encoding from content-type" in caplog.text


async def test_fetch_url_by_browser_errors(tmp_path, caplog):
    test_url = "http://test.com/error"
    output_dir = tmp_path / "output_error"
    config = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(output_dir),
        browser_delay=0,
    )

    # Mock Playwright objects
    mock_page = AsyncMock()
    mock_context = AsyncMock()
    mock_context.new_page.return_value = mock_page
    # Test TimeoutError
    mock_page.goto.side_effect = PlaywrightTimeoutError("Page timed out")
    with caplog.at_level(logging.ERROR):
        await fetch_url_by_browser(mock_context, test_url, config)
        assert (
            "Error fetching URL "
            "http://test.com/error with browser: Page timed out" in caplog.text
        )
    mock_page.close.assert_called_once()
    mock_page.close.reset_mock()

    # Test None response
    mock_page.goto.side_effect = None
    mock_page.goto.return_value = None
    with caplog.at_level(logging.ERROR):
        await fetch_url_by_browser(mock_context, test_url, config)
        assert (
            "Failed to get response for URL: http://test.com/error"
            in caplog.text
        )
    mock_page.close.assert_called_once()
    mock_page.close.reset_mock()

    # Test LookupError fallback
    test_url_lookup = "http://test.com/lookup"
    config_lookup = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(tmp_path / "output_lookup"),
        browser_delay=0,
    )
    mock_response_lookup = AsyncMock()
    mock_response_lookup.header_value.return_value = (
        "text/html; charset=invalid-encoding-name"
    )
    mock_page_lookup = AsyncMock()
    mock_page_lookup.goto.return_value = mock_response_lookup
    mock_page_lookup.content.return_value = "content"
    mock_context_lookup = AsyncMock()
    mock_context_lookup.new_page.return_value = mock_page_lookup
    with caplog.at_level(logging.INFO):
        await fetch_url_by_browser(
            mock_context_lookup,
            test_url_lookup,
            config_lookup,
        )
        assert "Unknown encoding 'invalid-encoding-name'" in caplog.text
        assert "Saved" in caplog.text

    expected_file_lookup = tmp_path / "output_lookup" / "lookup.html"
    assert expected_file_lookup.exists()
    assert expected_file_lookup.read_text(encoding="utf-8") == "content"


@patch("xsget.xsget._main_async_logic", new_callable=AsyncMock)
def test_run(mock_main_async_logic):
    config = argparse.Namespace(debug=True)
    run(config)
    mock_main_async_logic.assert_called_once_with(config)


@patch("xsget.xsget.run")
@patch("xsget.xsget.load_or_create_config", return_value={})
@patch("xsget.xsget.setup_logging")
@patch("xsget.xsget.build_parser")
def test_main_success(
    mock_build_parser,
    mock_setup_logging,
    mock_load_or_create_config,
    mock_run,
):
    mock_parser = MagicMock()
    mock_parser.parse_args.return_value = argparse.Namespace(
        url="http://test.com",
        debug=False,
    )
    mock_build_parser.return_value = mock_parser

    main(["http://test.com"])

    mock_setup_logging.assert_called_once()
    mock_load_or_create_config.assert_called_once()
    mock_run.assert_called_once()


@patch("xsget.xsget.run")
@patch(
    "xsget.xsget.load_or_create_config",
    side_effect=Exception("Config Error"),
)
@patch("xsget.xsget.setup_logging")
@patch("xsget.xsget.build_parser")
def test_main_error_handling(
    mock_build_parser,
    mock_setup_logging,
    mock_load_or_create_config,
    mock_run,
    caplog,
):
    mock_parser = MagicMock()
    mock_parser.parse_args.return_value = argparse.Namespace(
        url="http://test.com",
        debug=False,
    )
    mock_build_parser.return_value = mock_parser

    with pytest.raises(SystemExit) as excinfo, caplog.at_level(logging.ERROR):
        main(["http://test.com"])

    assert excinfo.value.code == 1
    assert "error: Config Error" in caplog.text
    mock_setup_logging.assert_called_once()
    mock_load_or_create_config.assert_called_once()
    mock_run.assert_not_called()


@patch("xsget.xsget.run")
@patch(
    "xsget.xsget.load_or_create_config",
    side_effect=Exception("Debug Error"),
)
@patch("xsget.xsget.setup_logging")
@patch("xsget.xsget.build_parser")
def test_main_error_handling_debug(
    mock_build_parser,
    mock_setup_logging,
    mock_load_or_create_config,
    mock_run,
    caplog,
):
    mock_parser = MagicMock()
    mock_parser.parse_args.return_value = argparse.Namespace(
        url="http://test.com",
        debug=True,
    )
    mock_build_parser.return_value = mock_parser

    with pytest.raises(SystemExit) as excinfo, caplog.at_level(logging.ERROR):
        main(["http://test.com", "-d"])

    assert excinfo.value.code == 1
    assert "error: Debug Error" in caplog.text
    mock_setup_logging.assert_called_once()
    mock_load_or_create_config.assert_called_once()
    mock_run.assert_not_called()


@patch("xsget.xsget.fetch_url_by_aiohttp", new_callable=AsyncMock)
async def test_fetch_urls_aiohttp(mock_fetch_aiohttp):
    urls = [["http://a.com/1", "http://a.com/2"]]
    config = argparse.Namespace(browser=False)

    await fetch_urls(urls, config)

    assert mock_fetch_aiohttp.call_count == 2
    mock_fetch_aiohttp.assert_any_call(ANY, "http://a.com/1", config)
    mock_fetch_aiohttp.assert_any_call(ANY, "http://a.com/2", config)


@patch("xsget.xsget.fetch_url_by_browser", new_callable=AsyncMock)
@patch("xsget.xsget.async_playwright")
async def test_fetch_urls_browser(mock_playwright, mock_fetch_browser):
    mock_browser = AsyncMock()
    mock_context = AsyncMock()
    mock_playwright.return_value.__aenter__.return_value.chromium.launch.return_value = mock_browser  # noqa: E501
    mock_browser.new_context.return_value = mock_context

    urls = [["http://b.com/1", "http://b.com/2"], ["http://b.com/3"]]
    config = argparse.Namespace(browser=True, browser_session=2)

    await fetch_urls(urls, config)

    mock_playwright.assert_called_once()
    mock_browser.new_context.assert_called_once()
    mock_browser.close.assert_called_once()
    mock_context.close.assert_called_once()

    assert mock_fetch_browser.call_count == 3
    mock_fetch_browser.assert_any_call(mock_context, "http://b.com/1", config)
    mock_fetch_browser.assert_any_call(mock_context, "http://b.com/2", config)
    mock_fetch_browser.assert_any_call(mock_context, "http://b.com/3", config)


@patch("xsget.xsget.extract_urls")
def test_filter_urls(mock_extract_urls, tmp_path, caplog):
    index_file = tmp_path / "index.html"
    index_file.write_text("<html>index</html>", encoding="utf-8")
    output_dir = tmp_path / "output"
    output_dir.mkdir()

    all_urls = [
        "http://a.com/new",
        "http://a.com/existing",
        "http://a.com/undecodable",
    ]
    mock_extract_urls.side_effect = [all_urls, all_urls, []]

    existing_filename = url_to_filename("http://a.com/existing")
    (output_dir / existing_filename).touch()

    config = argparse.Namespace(
        url_param_as_filename="",
        output_dir=str(output_dir),
        browser=False,
        browser_session=0,
    )

    with caplog.at_level(logging.INFO):
        burls = filter_urls(index_file, config)

    expected_urls = ["http://a.com/new", "http://a.com/undecodable"]
    assert burls == [expected_urls]
    assert "Total URL extracted: 3" in caplog.text
    assert "Total URL to download: 2" in caplog.text
    assert "Total URL batches to download: 1" in caplog.text

    config.browser = True
    config.browser_session = 1
    caplog.clear()
    with caplog.at_level(logging.INFO):
        burls = filter_urls(index_file, config)
    assert burls == [["http://a.com/new"], ["http://a.com/undecodable"]]
    assert "Total URL batches to download: 2" in caplog.text

    index_file_empty = tmp_path / "empty.html"
    index_file_empty.write_bytes(b"")
    config.browser = False
    config.browser_session = 0
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        burls = filter_urls(index_file_empty, config)
        assert burls == []


@patch("xsget.xsget.fetch_urls", new_callable=AsyncMock)
@patch("xsget.xsget.filter_urls")
def test_main_async_logic_range(mock_filter_urls, mock_fetch_urls):
    config = argparse.Namespace(
        url="http://test.com/page[1-3].html",
        url_param_as_filename="",
        output_dir="output",
        test=False,
    )
    asyncio.run(_main_async_logic(config))

    expected_urls = [
        "http://test.com/page1.html",
        "http://test.com/page2.html",
        "http://test.com/page3.html",
    ]
    mock_fetch_urls.assert_called_once_with([expected_urls], config)
    mock_filter_urls.assert_not_called()


@patch("xsget.xsget.fetch_urls", new_callable=AsyncMock)
@patch("xsget.xsget.filter_urls")
def test_main_async_logic_range_invalid(mock_filter_urls, mock_fetch_urls):
    config = argparse.Namespace(
        url="http://test.com/page[3-1].html",
        url_param_as_filename="",
        output_dir="output",
        test=False,
    )
    with pytest.raises(
        RuntimeError,
        match="invalid url range, start: 3, end: 1",
    ):
        asyncio.run(_main_async_logic(config))

    mock_fetch_urls.assert_not_called()
    mock_filter_urls.assert_not_called()


@patch("xsget.xsget.fetch_urls", new_callable=AsyncMock)
@patch("xsget.xsget.filter_urls")
def test_main_async_logic_single_url_crawl(
    mock_filter_urls,
    mock_fetch_urls,
    tmp_path,
):
    test_url = "http://test.com/index"
    output_dir = tmp_path / "output"
    output_dir.mkdir()
    index_filename = url_to_filename(test_url)
    index_path = output_dir / index_filename
    index_path.touch()

    config = argparse.Namespace(
        url=test_url,
        url_param_as_filename="",
        output_dir=str(output_dir),
        refresh=False,
        test=False,
    )

    mock_filter_urls.return_value = [["http://test.com/chap1"]]

    asyncio.run(_main_async_logic(config))

    mock_fetch_urls.assert_any_call([[test_url]], config)
    mock_filter_urls.assert_called_once_with(index_path, config)
    mock_fetch_urls.assert_any_call([["http://test.com/chap1"]], config)
    assert mock_fetch_urls.call_count == 2


@patch("xsget.xsget.fetch_urls", new_callable=AsyncMock)
@patch("xsget.xsget.filter_urls")
def test_main_async_logic_single_url_refresh(
    mock_filter_urls,
    mock_fetch_urls,
    tmp_path,
    caplog,
):
    test_url = "http://test.com/index"
    output_dir = tmp_path / "output"
    output_dir.mkdir()
    index_filename = url_to_filename(test_url)
    index_path = output_dir / index_filename
    index_path.touch()

    config = argparse.Namespace(
        url=test_url,
        url_param_as_filename="",
        output_dir=str(output_dir),
        refresh=True,
        test=False,
    )

    mock_filter_urls.return_value = [["http://test.com/chap1"]]

    with caplog.at_level(logging.INFO):
        asyncio.run(_main_async_logic(config))

    assert not index_path.exists()
    assert "Refresh the index url: http://test.com/index" in caplog.text
    mock_fetch_urls.assert_any_call([[test_url]], config)
    mock_filter_urls.assert_called_once_with(index_path, config)
    mock_fetch_urls.assert_any_call([["http://test.com/chap1"]], config)
    assert mock_fetch_urls.call_count == 2


@patch("xsget.xsget.fetch_urls", new_callable=AsyncMock)
@patch("xsget.xsget.filter_urls")
def test_main_async_logic_test_mode(
    mock_filter_urls,
    mock_fetch_urls,
    tmp_path,
    caplog,
):
    test_url = "http://test.com/index"
    output_dir = tmp_path / "output"
    output_dir.mkdir()
    index_filename = url_to_filename(test_url)
    index_path = output_dir / index_filename
    index_path.touch()

    config = argparse.Namespace(
        url=test_url,
        url_param_as_filename="",
        output_dir=str(output_dir),
        refresh=False,
        test=True,
    )

    mock_filter_urls.return_value = [
        ["http://test.com/chap1", "http://test.com/chap2"],
    ]

    with caplog.at_level(logging.INFO):
        asyncio.run(_main_async_logic(config))

    mock_fetch_urls.assert_called_once_with([[test_url]], config)
    mock_filter_urls.assert_called_once_with(index_path, config)
    assert "Found url: http://test.com/chap1" in caplog.text
    assert "Found url: http://test.com/chap2" in caplog.text
    assert mock_fetch_urls.call_count == 1


def test_url_to_filename_param_not_found():
    url = "http://example.com/?id=123"
    assert url_to_filename(url, "page") == "index.html"


def test_extract_urls_no_href():
    html = "<a>No Href</a>"
    config = argparse.Namespace(link_css_path="a", url="http://base.com")
    urls = extract_urls(html, config)
    assert urls == []


@pytest.mark.asyncio
async def test_fetch_url_by_aiohttp_oserror(mocker, caplog):
    config = argparse.Namespace(url_param_as_filename="", output_dir="out")
    session = MagicMock()
    resp = MagicMock()
    resp.url = "http://example.com"
    resp.charset = "utf-8"
    resp.text = mocker.AsyncMock(return_value="content")
    resp.raise_for_status = MagicMock()
    session.get.return_value.__aenter__.return_value = resp

    mocker.patch("aiofiles.open", side_effect=OSError("Disk full"))

    await fetch_url_by_aiohttp(session, "http://example.com", config)

    assert "Error writing to file" in caplog.text


@pytest.mark.asyncio
async def test_fetch_url_by_browser_no_charset(mocker):
    context = MagicMock()
    page = MagicMock()
    context.new_page = mocker.AsyncMock(return_value=page)

    response = MagicMock()
    response.header_value = mocker.AsyncMock(return_value="text/html")

    page.goto = mocker.AsyncMock(return_value=response)
    page.content = mocker.AsyncMock(return_value="html")
    page.wait_for_timeout = mocker.AsyncMock()
    page.close = mocker.AsyncMock()

    config = argparse.Namespace(
        browser_delay=0, url_param_as_filename="", output_dir="out"
    )

    mocker.patch("aiofiles.open")

    await fetch_url_by_browser(context, "http://example.com", config)


@pytest.mark.asyncio
async def test_fetch_url_by_browser_oserror(mocker, caplog):
    context = MagicMock()
    page = MagicMock()
    context.new_page = mocker.AsyncMock(return_value=page)

    response = MagicMock()
    response.header_value = mocker.AsyncMock(return_value="text/html")

    page.goto = mocker.AsyncMock(return_value=response)
    page.content = mocker.AsyncMock(return_value="html")
    page.wait_for_timeout = mocker.AsyncMock()
    page.close = mocker.AsyncMock()

    config = argparse.Namespace(
        browser_delay=0, url_param_as_filename="", output_dir="out"
    )

    mocker.patch("aiofiles.open", side_effect=OSError("Disk error"))

    await fetch_url_by_browser(context, "http://example.com", config)
    assert "Error writing to file" in caplog.text


@pytest.mark.asyncio
async def test_fetch_url_by_browser_fallback_error(mocker, caplog):
    context = MagicMock()
    page = MagicMock()
    context.new_page = mocker.AsyncMock(return_value=page)

    response = MagicMock()
    response.header_value = mocker.AsyncMock(
        return_value="text/html; charset=unknown"
    )

    page.goto = mocker.AsyncMock(return_value=response)
    page.content = mocker.AsyncMock(return_value="html")
    page.wait_for_timeout = mocker.AsyncMock()
    page.close = mocker.AsyncMock()

    config = argparse.Namespace(
        browser_delay=0, url_param_as_filename="", output_dir="out"
    )

    def side_effect(*args, **kwargs):
        if kwargs.get("encoding") == "unknown":
            raise LookupError("unknown encoding")
        if kwargs.get("encoding") == "utf-8":
            raise OSError("fallback failed")
        return mocker.AsyncMock()

    mocker.patch("aiofiles.open", side_effect=side_effect)

    await fetch_url_by_browser(context, "http://example.com", config)
    assert "Failed to write" in caplog.text


@pytest.mark.asyncio
async def test_fetch_url_by_browser_page_none(mocker, caplog):
    context = MagicMock()
    context.new_page = mocker.AsyncMock(
        side_effect=PlaywrightError("Playwright Error")
    )

    config = argparse.Namespace(browser_delay=0)

    await fetch_url_by_browser(context, "http://example.com", config)

    assert "Error fetching URL" in caplog.text


def test_filter_urls_undecodable(mocker, caplog):
    index_html = Path("index.html")
    config = argparse.Namespace()

    mocker.patch("pathlib.Path.open", mocker.mock_open(read_data=b""))

    with patch("xsget.xsget.UnicodeDammit") as MockDammit:
        MockDammit.return_value.unicode_markup = None

        res = filter_urls(index_html, config)
        assert res == []
        assert "Skipping empty or undecodable file" in caplog.text


@pytest.mark.asyncio
async def test_main_async_logic_refresh_no_file(mocker, caplog):
    caplog.set_level(logging.INFO)
    config = argparse.Namespace(
        url="http://example.com",
        refresh=True,
        output_dir="out",
        url_param_as_filename="",
        test=False,
        browser=False,
    )

    mocker.patch("xsget.xsget.fetch_urls")
    mocker.patch("xsget.xsget.filter_urls", return_value=[])

    await _main_async_logic(config)

    assert "Refresh the index url" in caplog.text



# Copyright (C) 2021,2022,2023,2024,2025,2026 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=C0114,C0116

import argparse
import logging
import os
import shutil
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
import regex as re
from bs4 import BeautifulSoup

from xsget.chapter import Chapter
from xsget.xstxt import (
    DEFAULT_CONFIG_FILE,
    ConfigFileHandler,
    _compile_regex_replacements,
    _read_and_extract_chapter,
    _run_monitor,
    extract_body,
    extract_chapter,
    extract_title,
    generate_book,
    generate_multiple_txt,
    generate_single_txt,
    get_html_files,
    main,
    search_and_replace,
    wrap,
)

# Taken from: https://zh.wikisource.org/wiki/詩經/關雎
CTEXT = """\
　　孔子論《詩》，以《關雎》為始。言太上者民之父母，后夫人之
行不侔乎天地，則無以奉神靈之统而理萬物之宜，故《詩》曰：「窈
窕淑女，君子好逑。」言能致其貞淑，不贰其操，情欲之感無介乎容
儀，晏私之意不形乎動静，夫然後可以配至尊而為宗廟主。此綱紀之
首、王教之端也。
"""


def test_get_html_files_sorted_in_natural_order(tmpdir):
    for name in ["200", "2", "100"]:
        tmpdir.join(f"{name}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 0, []) == [
        Path(f"{path}/2.html"),
        Path(f"{path}/100.html"),
        Path(f"{path}/200.html"),
    ]


def test_get_html_files_exclude_filter(tmpdir):
    for name in ["a", "2", "xyz"]:
        tmpdir.join(f"{name}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 0, [f"{path}/xyz.html"]) == [
        Path(f"{path}/2.html"),
        Path(f"{path}/a.html"),
    ]


def test_get_single_html_file(tmpdir):
    single = tmpdir.join("single.html")
    single.write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/single.html"], 0, []) == [
        Path(f"{path}/single.html"),
    ]


def test_get_html_files_by_limit(tmpdir):
    for i in range(5):
        tmpdir.join(f"{i}.html").write("")

    path = str(tmpdir)
    assert get_html_files([f"{path}/*.html"], 3, []) == [
        Path(f"{path}/0.html"),
        Path(f"{path}/1.html"),
        Path(f"{path}/2.html"),
    ]


def test_get_no_html_files(caplog):
    path = "foobar/*.html"
    assert get_html_files([path], 0, []) == []
    assert f"No input files found in: {path}" in caplog.text


def test_get_html_files_with_non_existent_exclude(caplog):
    # Create some dummy files
    path = "temp_dir"
    Path(path).mkdir(exist_ok=True)
    Path(path, "file1.html").touch()
    Path(path, "file2.html").touch()

    # Call with a non-existent exclude pattern
    result = get_html_files(
        [f"{path}/*.html"],
        0,
        [f"{path}/non_existent_*.html"],
    )
    assert len(result) == 2  # Both files should still be included
    assert (
        f"No exclude files found in: {path}/non_existent_*.html" in caplog.text
    )

    # Clean up
    shutil.rmtree(path)


async def test_extract_chapter():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">My Content</div>
        </body>
        </html>
    """
    config = argparse.Namespace(
        title_css_path="title",
        body_css_path="div#content",
        html_replace=[],
        compiled_html_replace=[],
        debug=False,
    )
    chapter = await extract_chapter(html, config)
    chapter.filename = "123.html"
    assert chapter.filename == "123.html"
    assert chapter.title == "My Title"
    assert chapter.content == "My Content"
    assert str(chapter) == "My Title\n\nMy Content"
    assert repr(chapter) == (
        "Chapter(content='My Content', title='My Title', "
        "filename='123.html', content_path='')"
    )


@pytest.mark.asyncio
async def test_extract_chapter_with_html_replace():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph1
            <br/><br />
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph2
            <br/>
            <br />
            &nbsp;&nbsp;&nbsp;&nbsp;Paragraph3
            </div>
        </body>
        </html>
    """
    raw_replacements = [
        ("<br/>", "11"),
        ("<br />", "22"),
        ("&nbsp;&nbsp;", "33"),
    ]
    compiled_replacements = [
        (re.compile(r"<br/>", re.MULTILINE), r"11"),
        (re.compile(r"<br />", re.MULTILINE), r"22"),
        (re.compile(r"&nbsp;&nbsp;", re.MULTILINE), r"33"),
    ]
    config = argparse.Namespace(
        title_css_path="title",
        body_css_path="div#content",
        html_replace=raw_replacements,
        compiled_html_replace=compiled_replacements,
        debug=False,
    )
    chapter = await extract_chapter(html, config)

    match_regex = (
        "My Title\n\n"
        r"\n\s+3333Paragraph1"
        r"\n\s+1122"
        r"\n\s+3333Paragraph2"
        r"\n\s+11"
        r"\n\s+22"
        r"\n\s+3333Paragraph3"
    )
    assert re.match(match_regex, str(chapter))


async def test_extract_chapter_without_css_path():
    html = """
        <html>
        <head><title>My Title</title></head>
        <body>
            <div id="content">My Content</div>
        </body>
        </html>
    """
    config = argparse.Namespace(
        title_css_path=None,
        body_css_path=None,
        html_replace=[],
        compiled_html_replace=[],
        debug=False,
    )
    chapter = await extract_chapter(html, config)
    assert chapter.title == ""
    assert chapter.content == ""
    assert str(chapter) == ""
    assert (
        repr(chapter)
        == "Chapter(content='', title='', filename='', content_path='')"
    )


def test_generate_single_txt(tmpdir):
    chapters = [
        Chapter("MyTitle1", "MyContent1"),
        Chapter("MyTitle2", "MyContent2"),
    ]

    config = argparse.Namespace(
        book_title="Book Title",
        book_author="Book Author",
        output=str(Path(tmpdir, "book.txt")),
        txt_replace=(),
        compiled_txt_replace=[],
        width=60,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
        language="zh",
        output_dir="output",
        gettext_func=lambda x: x,
    )
    generate_single_txt(chapters, config)

    with Path(config.output).open(encoding="utf8") as file:
        content = file.read()
        assert content == (
            "---\n"
            "书名：Book Title\n"
            "作者：Book Author\n"
            "---\n\n"
            "MyTitle1\n\nMyContent1\n\n"
            "MyTitle2\n\nMyContent2"
        )


def test_generate_single_txt_with_search_and_replace(tmpdir):
    book = [
        Chapter("MyTitle1", "MyContent1"),
        Chapter("MyTitle2", "MyContent2"),
    ]

    raw_replacements = [
        ("Title", "TITLE"),
        ("My", "YY"),
    ]
    compiled_replacements = [
        (re.compile(r"Title", re.MULTILINE), r"TITLE"),
        (re.compile(r"My", re.MULTILINE), r"YY"),
    ]
    config = argparse.Namespace(
        book_title="Book Title",
        book_author="Book Author",
        output=str(Path(tmpdir, "book.txt")),
        txt_replace=raw_replacements,
        compiled_txt_replace=compiled_replacements,
        width=60,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
        language="zh",
        output_dir="output",
        gettext_func=lambda x: x,
    )
    generate_single_txt(book, config)

    with Path(config.output).open(encoding="utf8") as file:
        content = file.read()
        assert content == (
            "---\n"
            "书名：Book Title\n"
            "作者：Book Author\n"
            "---\n\n"
            "YYTITLE1\n\n"
            "YYContent1\n\n"
            "YYTITLE2\n\n"
            "YYContent2"
        )


def test_chapter_model_to_str():
    fixture_and_result = [
        (Chapter("MyTitle", "MyContent"), "MyTitle\n\nMyContent"),
        (Chapter("MyTitle", ""), "MyTitle"),
        (Chapter(title="MyTitle"), "MyTitle"),
        (Chapter("", "MyContent"), "MyContent"),
        (Chapter(content="MyContent"), "MyContent"),
        (Chapter(), ""),
    ]
    for fixture, result in fixture_and_result:
        assert str(fixture) == result


def test_compile_regex_replacements_with_invalid_regex():
    with pytest.raises(re.error):
        _compile_regex_replacements([("[", "")])


def test_search_and_replace_with_empty_regexs():
    content = "This is some content."
    result = search_and_replace(content, [])
    assert result == content


def test_wrap_by_width():
    config = argparse.Namespace(
        width=20,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert first_line == "\u3000\u3000孔子論《詩》，以"


def test_wrap_with_empty_content():
    config = argparse.Namespace(
        width=20,
        indent_chars="",
        paragraph_separator="\n\n",
        fullwidth=False,
    )
    assert wrap("", config) == ""
    assert wrap("   ", config) == ""  # `wrap` strips whitespace for width > 0


def test_no_wrap():
    config = argparse.Namespace(
        width=0,
        indent_chars="",
        paragraph_separator="\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert first_line == (
        "\u3000\u3000孔子論《詩》，以《關雎》為始。言太上者民之父母，后夫人之"
    )


def test_wrap_with_indent_characters():
    config = argparse.Namespace(
        width=20,
        indent_chars="0000",
        paragraph_separator="\n",
        fullwidth=False,
    )
    first_line = wrap(CTEXT, config).split("\n", maxsplit=1)[0]
    assert first_line == "0000孔子論《詩》"


async def test_extract_title_no_match():
    html = "<html><head><title>My Title</title></head><body></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    title = extract_title(soup, "h1")  # Non-existent CSS path
    assert title == ""


async def test_extract_title_empty_element():
    html = "<html><head><title></title></head><body></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    title = extract_title(
        soup,
        "title",
    )  # Existing CSS path, but empty element
    assert title == ""


async def test_extract_body_no_match():
    html = "<html><body><p>Content</p></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    body = extract_body(soup, "div#content")  # Non-existent CSS path
    assert body == ""


async def test_extract_body_empty_element():
    html = "<html><body><div id='content'></div></body></html>"
    soup = BeautifulSoup(html, features="lxml")
    body = extract_body(
        soup,
        "div#content",
    )  # Existing CSS path, but empty element
    assert body == ""


def test_generate_multiple_txt(tmpdir):
    chapters = [
        Chapter("Title1", "Content1", filename="file1.html"),
        Chapter("Title2", "Content2", filename="file2.html"),
    ]
    output_dir = Path(tmpdir, "output_files")
    config = argparse.Namespace(
        output_dir=str(output_dir),
        txt_replace=[],
        compiled_txt_replace=[],
        fullwidth=False,
        width=0,
        indent_chars="",
        paragraph_separator="\n\n",
        language="en",
    )
    generate_multiple_txt(chapters, config)

    assert Path(output_dir, "file1.txt").exists()
    assert Path(output_dir, "file2.txt").exists()
    with Path(output_dir, "file1.txt").open(encoding="utf8") as f:
        assert f.read() == "Title1\n\nContent1"
    with Path(output_dir, "file2.txt").open(encoding="utf8") as f:
        assert f.read() == "Title2\n\nContent2"


@pytest.mark.asyncio
async def test_generate_book_purge(mocker, tmpdir):
    # Mock necessary functions to avoid actual file operations and async calls
    mocker.patch("xsget.xstxt.get_html_files", return_value=["dummy.html"])
    mocker.patch("xsget.xstxt.extract_chapter", return_value=Chapter("T", "C"))
    mocker.patch("xsget.xstxt.generate_single_txt")
    mocker.patch("xsget.xstxt.generate_multiple_txt")
    mocker.patch("shutil.rmtree")  # Mock rmtree

    # Mock aiofiles.open to prevent FileNotFoundError and support async with
    # Create a mock for the file handle that will be returned by aiofiles.open
    mock_file_handle = mocker.AsyncMock()
    mock_file_handle.read.return_value = b"<html><body>Content</body></html>"

    # Explicitly set __aenter__ and __aexit__ for the mock_file_handle
    # This is often not needed for AsyncMock, but can help debug stubborn
    # issues.
    mock_file_handle.__aenter__ = mocker.AsyncMock(
        return_value=mock_file_handle,
    )
    mock_file_handle.__aexit__ = mocker.AsyncMock(return_value=None)

    # Patch aiofiles.open to return this mock_file_handle directly.
    mocker.patch("aiofiles.open", return_value=mock_file_handle)

    output_dir = Path(tmpdir, "output")
    output_dir.mkdir()

    config = argparse.Namespace(
        input=["*.html"],
        limit=0,
        exclude=[],
        debug=False,
        output_dir=str(output_dir),
        output="book.txt",
        purge=True,
        yes=False,
        output_individual_file=False,
        html_replace=[],
        txt_replace=[],
        compiled_html_replace=[],
        compiled_txt_replace=[],
    )

    await generate_book(config)

    # Assert that rmtree was called with the correct path
    shutil.rmtree.assert_called_once_with(output_dir)


@pytest.mark.asyncio
async def test_read_and_extract_chapter_debug(mocker, caplog):
    caplog.set_level(logging.DEBUG)
    config = argparse.Namespace(debug=True)
    filename = Path("test.html")

    # Mock aiofiles to return content so it proceeds to completion or error
    mock_file = mocker.AsyncMock()
    # Return bytes that are valid
    mock_file.read.return_value = b"<html></html>"
    mock_file.__aenter__.return_value = mock_file
    mocker.patch("aiofiles.open", return_value=mock_file)
    mocker.patch("xsget.xstxt.extract_chapter", return_value=Chapter("T", "C"))

    await _read_and_extract_chapter(filename, config, 1, 10)

    assert "Processing file: test.html" in caplog.text


@pytest.mark.asyncio
async def test_read_and_extract_chapter_progress(mocker, capsys):
    config = argparse.Namespace(debug=False)
    filename = Path("test.html")

    mocker.patch("sys.stdout.isatty", return_value=True)

    mock_file = mocker.AsyncMock()
    mock_file.read.return_value = b"<html></html>"
    mock_file.__aenter__.return_value = mock_file
    mocker.patch("aiofiles.open", return_value=mock_file)
    mocker.patch("xsget.xstxt.extract_chapter", return_value=Chapter("T", "C"))

    await _read_and_extract_chapter(filename, config, 5, 10)

    captured = capsys.readouterr()
    assert "Processing file: test.html (50.0% - 5 / 10)" in captured.out


@pytest.mark.asyncio
async def test_extract_chapter_debug(caplog):
    caplog.set_level(logging.DEBUG)
    config = argparse.Namespace(
        debug=True,
        compiled_html_replace=[],
        title_css_path="h1",
        body_css_path="p",
    )
    from bs4 import BeautifulSoup

    soup = BeautifulSoup("<h1>T</h1><p>C</p>", "html.parser")

    await extract_chapter(soup, config, "test.html")

    assert (
        "Processing Chapter(content='C', title='T', filename='test.html', content_path='')"
        in caplog.text
    )


@pytest.mark.asyncio
async def test_extract_chapter_progress(mocker, capsys):
    config = argparse.Namespace(
        debug=False,
        compiled_html_replace=[],
        title_css_path="h1",
        body_css_path="p",
    )
    mocker.patch("sys.stdout.isatty", return_value=True)
    from bs4 import BeautifulSoup

    soup = BeautifulSoup("<h1>T</h1><p>C</p>", "html.parser")

    await extract_chapter(soup, config, "test.html")

    captured = capsys.readouterr()
    assert "Processing Chapter" in captured.out


def test_wrap_negative_width():
    config = argparse.Namespace(
        width=-1, paragraph_separator="\n", indent_chars="  "
    )
    content = "Hello World"
    # width < 0 means infinite width (sys.maxsize) when indenting
    res = wrap(content, config)
    assert res == "  Hello World"


def test_generate_multiple_txt_options(tmp_path, caplog):
    caplog.set_level(logging.INFO)
    chapters = [Chapter("T", "Content", filename="ch1.html")]
    output_dir = tmp_path / "out"
    config = argparse.Namespace(
        output_dir=str(output_dir),
        output_individual_file=True,
        compiled_txt_replace=[(re.compile("Content"), "NewContent")],
        fullwidth=True,
        width=0,
        paragraph_separator="\n",
        indent_chars="",
        txt_replace=[],
    )

    generate_multiple_txt(chapters, config)

    assert (output_dir / "ch1.txt").exists()
    assert "Converting halfwidth ASCII to fullwidth" in caplog.text
    content = (output_dir / "ch1.txt").read_text(encoding="utf8")


def test_generate_single_txt_options(tmp_path, caplog):
    caplog.set_level(logging.INFO)
    chapters = [Chapter("T", "Content")]
    config = argparse.Namespace(
        output_dir=str(tmp_path),
        output="book.txt",
        compiled_txt_replace=[],
        fullwidth=True,
        width=0,
        paragraph_separator="\n",
        indent_chars="",
        book_title="BT",
        book_author="BA",
        gettext_func=lambda x: x,
    )

    generate_single_txt(chapters, config)
    assert "Converting halfwidth ASCII to fullwidth" in caplog.text


def test_main_exception(mocker, caplog):
    mocker.patch(
        "xsget.xstxt._load_config", side_effect=ValueError("Test Error")
    )

    with pytest.raises(SystemExit):
        main([])

    assert "error: Test Error" in caplog.text


@pytest.mark.asyncio
async def test_read_and_extract_chapter_undecodable(mocker, caplog):
    config = argparse.Namespace(debug=False)
    filename = Path("bad.html")

    mock_file = mocker.AsyncMock()
    mock_file.read.return_value = b"\xff"
    mock_file.__aenter__.return_value = mock_file

    mocker.patch("aiofiles.open", return_value=mock_file)

    # Mock UnicodeDammit to return None
    with patch("xsget.xstxt.UnicodeDammit") as MockDammit:
        instance = MockDammit.return_value
        instance.unicode_markup = None

        result = await _read_and_extract_chapter(filename, config, 1, 1)
        assert result is None
        assert "Skipping empty or undecodable file" in caplog.text


@pytest.mark.asyncio
async def test_generate_book_output_individual(mocker):
    config = argparse.Namespace(
        input=["*.html"],
        limit=0,
        exclude=[],
        debug=False,
        output_dir="out",
        purge=False,
        output_individual_file=True,
        yes=False,
    )

    mocker.patch("xsget.xstxt.get_html_files", return_value=[Path("1.html")])
    mocker.patch(
        "xsget.xstxt._read_and_extract_chapter",
        return_value=Chapter("T", "C", filename="1.txt"),
    )
    mock_gen_multi = mocker.patch("xsget.xstxt.generate_multiple_txt")

    mocker.patch("sys.stdout.isatty", return_value=True)
    mocker.patch(
        "os.get_terminal_size", return_value=os.terminal_size((80, 24))
    )

    await generate_book(config)

    mock_gen_multi.assert_called_once()


def test_run_monitor(mocker):
    config = argparse.Namespace(config="config.toml")

    mock_observer = mocker.patch("xsget.xstxt.Observer")
    mock_observer_instance = mock_observer.return_value
    mock_observer_instance.is_alive.side_effect = [True, False]

    _run_monitor(config)

    mock_observer_instance.start.assert_called_once()
    mock_observer_instance.stop.assert_called_once()


def test_config_file_handler(mocker):
    # Important: _load_config returns a Namespace.
    # self.config.config is accessed in on_modified.
    config = argparse.Namespace(config="test_config.toml", debug=False)
    handler = ConfigFileHandler(config)

    event = MagicMock()
    event.src_path = DEFAULT_CONFIG_FILE  # "xstxt.toml"

    mocker.patch("xsget.xstxt._load_config", return_value=config)
    mocker.patch("xsget.xstxt.generate_book")

    handler.on_modified(event)

    # Need to check call on the module function, but it's imported in this test file
    # `generate_book` is imported. `xstxt.generate_book` patch will work if
    # ConfigFileHandler calls it from the module.
    # In `src/xsget/xstxt.py`: `asyncio.run(generate_book(config), debug=config.debug)`
    # Since `ConfigFileHandler` is defined in `xstxt.py`, it sees `generate_book` in globals.

    # Let's verify the patch target. `mocker.patch("xsget.xstxt.generate_book")`
    # this mocks the function in the module. ConfigFileHandler uses that.

    # We can assert on the mock.
    # But wait, we didn't capture the mock object here.
    # mocker.patch returns the mock.
    mock_gen = mocker.patch("xsget.xstxt.generate_book")
    # Wait, I already patched it above but didn't assign.
    # Let's re-patch with variable.

    # Actually, in the previous code block I did:
    # mocker.patch("xsget.xstxt.generate_book")
    # handler.on_modified(event)
    # xstxt.generate_book.assert_called_once() -> this referred to the imported module `xstxt` in the temp file.
    # Here I don't have `xstxt` module imported entirely, I imported functions.

    # So I should change how I assert.
    # `xsget.xstxt.generate_book` is being patched. this replaces the object in sys.modules.
    # `from xsget.xstxt import generate_book` in this test file might still point to original if imported before patch?
    # Mocker handles this if setup correctly.

    # Safest way:
    mock_generate_book = mocker.patch("xsget.xstxt.generate_book")

    handler.on_modified(event)

    mock_generate_book.assert_called_once()


def test_main_monitor(mocker):
    config = argparse.Namespace(debug=False, monitor=True)
    mocker.patch("xsget.xstxt._load_config", return_value=config)
    mocker.patch("xsget.xstxt.generate_book")
    mock_run_monitor = mocker.patch("xsget.xstxt._run_monitor")

    main([])

    mock_run_monitor.assert_called_once()


def test_config_file_handler_ignore_other_files(mocker):
    config = argparse.Namespace()
    handler = ConfigFileHandler(config)

    event = MagicMock()
    event.src_path = "other.xml"

    # Needs to be mocked because it's imported in the module scope
    mock_load = mocker.patch("xsget.xstxt._load_config")

    handler.on_modified(event)

    mock_load.assert_not_called()



# Copyright (C) 2021,2022,2023,2024,2025,2026 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Console tools to download online novel and convert to text file."""

import logging
from importlib import metadata

from .config import (
    ConfigFileCorruptedError as ConfigFileCorruptedError,
)
from .config import (
    ConfigFileExistsError as ConfigFileExistsError,
)
from .config import (
    load_or_create_config as load_or_create_config,
)
from .config import (
    setup_logging as setup_logging,
)

__version__ = metadata.version("xsget")

_logger = logging.getLogger(__name__)



# Copyright (C) 2021,2022,2023,2024,2025,2026 Kian-Meng Ang
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""xstxt is a console app that extract content from HTML to text file.

website: https://github.com/kianmeng/xsget
changelog: https://github.com/kianmeng/xsget/blob/master/CHANGELOG.md
issues: https://github.com/kianmeng/xsget/issues
"""

import argparse
import asyncio
import gettext
import logging
import math
import os
import shutil
import sys
import textwrap
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import TypedDict

import aiofiles
import regex as re
from bs4 import BeautifulSoup, UnicodeDammit
from natsort import natsorted
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from xsget import (
    load_or_create_config,
    setup_logging,
)
from xsget.arg_parser import create_base_parser
from xsget.chapter import Chapter

__usages__ = """
examples:
  xsget -g
  xstxt --input *.html
  xstxt --output-individual-file --input *.html
  xstxt --config --monitor

"""

# Unicode integer in hexadecimal for these characters.
FULLWIDTH_EXCLAMATION_MARK = 0xFF01
EXCLAMATION_MARK = 0x21
TILDE = 0x7E

# Fullwidth is a text character that occupies two alphanumeric characters
# in monospace font.
#
# See Halfwidth and Fullwidth Forms in Unicode (https://w.wiki/66Ps) and
# Unicode block (https://w.wiki/66Pt).
HALFWIDTH_FULLWIDTH_MAP = {}
for halfwidth_i, fullwidth_i in enumerate(range(EXCLAMATION_MARK, TILDE + 1)):
    HALFWIDTH_FULLWIDTH_MAP[fullwidth_i] = (
        FULLWIDTH_EXCLAMATION_MARK + halfwidth_i
    )

DEFAULT_CONFIG_FILE = "xstxt.toml"

_logger = logging.getLogger(__name__)


class ConfigFileHandler(FileSystemEventHandler):  # type: ignore[misc]
    """Custom event handler for monitoring changes in config file."""

    def __init__(self, config: argparse.Namespace) -> None:
        """Initialize the config file handler.

        Args:
            config: Configuration from `xstxt.toml`

        """
        self.config = config

    def on_modified(self, event: FileSystemEvent) -> None:
        """Re-generate the book when config file updated."""
        if Path(event.src_path).name == DEFAULT_CONFIG_FILE:
            config = _load_config(["-c", self.config.config])
            asyncio.run(generate_book(config), debug=config.debug)


def get_html_files(
    inputs: list[str],
    limit: int,
    excludes: list[str],
) -> list[Path]:
    """Get the list of HTML files or file for cleansing and extracting.

    Args:
        inputs (list[str]): Glob-like pattern for selecting HTML files
        limit (int): Number of HTML files to process
        excludes (list[str]): Glob-like pattern for excluding HTML files

    Returns:
        list[Path]: Number of HTML file names
    """
    input_files = set()
    for pattern in inputs:
        found_files = {
            p
            for p in Path(pattern).parent.glob(Path(pattern).name)
            if p.is_file()
        }
        if not found_files:
            _logger.warning("No input files found in: %s", pattern)
        input_files.update(found_files)

    exclude_files = set()
    for pattern in excludes:
        found_files = {
            p
            for p in Path(pattern).parent.glob(Path(pattern).name)
            if p.is_file()
        }
        if not found_files:
            _logger.warning("No exclude files found in: %s", pattern)
        exclude_files.update(found_files)

    files: list[Path] = natsorted(list(input_files - exclude_files), key=str)
    return files[:limit] if limit > 0 else files


async def _read_and_extract_chapter(
    filename: Path,
    config: argparse.Namespace,
    i: int,
    total_files: int,
) -> Chapter | None:
    """Read HTML file, decode, and extract chapter content.

    Args:
        filename (Path): The path to the HTML file.
        config (argparse.Namespace): config from args or file.
        i (int): Current file index (for progress).
        total_files (int): Total number of files (for progress).

    Returns:
        Chapter | None: The extracted chapter or None if the file is
        empty/undecodable.
    """
    if config.debug:
        _logger.debug("Processing file: %s", filename)
    elif sys.stdout.isatty():
        percent = round(i / total_files * 100, 1)
        progress = f"({percent}% - {i} / {total_files})"
        print(
            f"Processing file: {filename} {progress}",
            end="\r",
            flush=True,
        )

    async with aiofiles.open(filename, "rb") as file:
        dammit = UnicodeDammit(await file.read())
        decoded_html = dammit.unicode_markup
        if decoded_html is None:
            _logger.warning("Skipping empty or undecodable file: %s", filename)
            return None
        soup = BeautifulSoup(decoded_html, features="lxml")
        return await extract_chapter(soup, config, str(filename))


async def generate_book(config: argparse.Namespace) -> None:
    """Extract all chapters from HTML files into single text file.

    Args:
        config (argparse.Namespace): config from args or file
    """
    html_files = get_html_files(config.input, config.limit, config.exclude)
    total_files = len(html_files)
    _logger.info("Processing total files: %d", total_files)

    futures = [
        _read_and_extract_chapter(filename, config, i, total_files)
        for i, filename in enumerate(html_files, start=1)
    ]

    chapters_with_none = await asyncio.gather(*futures)
    chapters = [c for c in chapters_with_none if c is not None]

    # Clear the progress line only if running in a TTY
    if sys.stdout.isatty():
        print(" " * os.get_terminal_size()[0], end="\r")

    output_dir_path = Path.cwd() / config.output_dir
    if config.purge and output_dir_path.exists() and not config.yes:
        _logger.debug("Purge output folder: %s", output_dir_path)
        shutil.rmtree(output_dir_path)

    if config.output_individual_file:
        generate_multiple_txt(chapters, config)
    else:
        generate_single_txt(chapters, config)


async def extract_chapter(
    decoded_soup: BeautifulSoup,
    config: argparse.Namespace,
    filename: str = "",
) -> Chapter:
    """Extract chapter from the decoded HTML.

    Args:
        decoded_soup (BeautifulSoup): decoded HTML text as BeautifulSoup object
        config (argparse.Namespace): config from args or file
        filename (str): the filename of the decoded html

    Returns:
        Chapter: extracted chapters
    """
    html_content = str(decoded_soup)
    if config.compiled_html_replace:
        html_content = search_and_replace(
            html_content,
            config.compiled_html_replace,
        )

    soup = BeautifulSoup(html_content, features="lxml")
    title = extract_title(soup, config.title_css_path)
    body = extract_body(soup, config.body_css_path)

    chapter = Chapter(title=title, content=body.rstrip(), filename=filename)

    if config.debug:
        _logger.debug("Processing %s", repr(chapter))
    elif sys.stdout.isatty():
        print(f"Processing {chapter!r}", end="\r")

    return chapter


def extract_title(decoded_html: BeautifulSoup, css_path: str | None) -> str:
    """Extract title of a chapter from HTML.

    Args:
        decoded_html (BeautifulSoup): HTML text
        css_path (str): CSS path to a title of a chapter

    Returns:
        str: title of a chapter
    """
    if not css_path:
        return ""

    title = decoded_html.select_one(css_path)
    return title.text.strip() if title else ""


def extract_body(html: BeautifulSoup, css_path: str) -> str:
    """Extract body of a chapter from HTML.

    Args:
        html (BeautifulSoup): HTML text
        css_path (str): CSS path to a body of a chapter

    Returns:
        str: The body of HTML page if found, otherwise empty string.
    """
    if not css_path:
        return ""

    body = html.select_one(css_path)
    return body.text if body else ""


def search_and_replace(
    content: str,
    compiled_regexs: list[tuple[re.Pattern[str], str]],
) -> str:
    """Replace words/phrases based on a list of compiled regex.

    Args:
        content (str): HTML or plain text
        compiled_regexs (list): List of compiled regex rules

    Returns:
        str: HTML or plain text
    """
    for before, after in compiled_regexs:
        content = re.sub(before, after, content)
    return content


def _calculate_wrap_width(content: str, configured_width: int) -> int:
    """Calculate the effective wrap width, adjusting for CJK characters.

    If the content contains CJK characters, the width is halved.

    Args:
        content (str): The text content to check.
        configured_width (int): The width set by the user.

    Returns:
        int: The calculated effective wrap width.
    """
    calculated_width = configured_width
    # Check for Unicode CJK Unified Ideographs code block (4E00—9FFF)
    if re.search(r"[\u4e00-\u9fff]+", content):
        calculated_width = math.floor(configured_width // 2)
    return calculated_width


class FillOptions(TypedDict, total=False):
    """Options for text wrapping using `textwrap.fill`."""

    width: int
    initial_indent: str


def wrap(content: str, config: argparse.Namespace) -> str:
    """Wrap the content to a length.

    We assume that each paragraph was separated by an empty line.

    config.width is the length of the line to wrap.

    If the content falls within Unicode CJK Unified Ideographs code block
    (4E00—9FFF), we treat it as multi-bytes character and divide the configured
    width by 2.

    And text wrapping for CJK text is rather complicated. See
    https://github.com/python/cpython/issues/68853.

    Args:
        content (str): HTML or plain text.
        config (argparse.Namespace): config from args or file

    Returns:
        str: HTML or plain text
    """
    if not content:
        return content

    options: FillOptions = {}

    if config.width > 0:
        calculated_width = _calculate_wrap_width(content, config.width)

        _logger.debug(
            "Wrap paragraph at width: calculated: %d, configured: %d",
            calculated_width,
            config.width,
        )
        options["width"] = calculated_width

    paragraphs = []
    # Assuming each paragraph was separated by an empty line based on default
    # value for `-ps` argument.
    for paragraph in content.split(config.paragraph_separator):
        processed_paragraph = paragraph
        if config.indent_chars != "":
            processed_paragraph = textwrap.dedent(processed_paragraph).strip()
            options["initial_indent"] = config.indent_chars
            # default `width` is 70 if not set and we need to set to larger
            # value instead.
            if config.width < 0:
                options["width"] = sys.maxsize

        if config.width > 0:
            processed_paragraph = processed_paragraph.rstrip().replace(
                "\n", ""
            )

        if options:
            _logger.debug(options)
            processed_paragraph = textwrap.fill(processed_paragraph, **options)

        paragraphs.append(processed_paragraph)

    wrapped_content = config.paragraph_separator.join(paragraphs)
    return str(wrapped_content)


def generate_multiple_txt(
    chapters: list[Chapter],
    config: argparse.Namespace,
) -> None:
    """Write the extracted book into multiple txt file.

    Args:
        chapters (list): A list of Chapter.
        config (argparse.Namespace): config from args or file
    """
    for chapter in chapters:
        content = str(chapter)
        if config.compiled_txt_replace:
            content = search_and_replace(content, config.compiled_txt_replace)

        if config.fullwidth:
            _logger.info("Converting halfwidth ASCII to fullwidth")
            content = content.translate(HALFWIDTH_FULLWIDTH_MAP)

        content = wrap(content, config)

        filename = (
            Path(config.output_dir) / Path(chapter.filename).stem
        ).with_suffix(".txt")
        filename.parent.mkdir(parents=True, exist_ok=True)
        with Path(filename).open("w", newline="\n", encoding="utf8") as file:
            file.write(content)


def generate_single_txt(
    chapters: list[Chapter],
    config: argparse.Namespace,
) -> None:
    """Write the extracted book into single txt file.

    Args:
        chapters (list): A list of Chapters
        config (argparse.Namespace): config from args or file
    """
    filename = Path(config.output_dir) / config.output
    filename.parent.mkdir(parents=True, exist_ok=True)
    with Path(filename).open("w", newline="\n", encoding="utf8") as file:
        _ = config.gettext_func
        file.write("---\n")
        file.write(_("书名：") + config.book_title + "\n")
        file.write(_("作者：") + config.book_author + "\n")
        file.write(f"---{config.paragraph_separator}")

        chapter_strings = [str(chapter) for chapter in chapters]
        content = config.paragraph_separator.join(chapter_strings)

        if config.compiled_txt_replace:
            content = search_and_replace(content, config.compiled_txt_replace)

        if config.fullwidth:
            _logger.info("Converting halfwidth ASCII to fullwidth")
            content = content.translate(HALFWIDTH_FULLWIDTH_MAP)

        content = wrap(content, config)

        file.write(content)


def build_parser() -> argparse.ArgumentParser:
    """Build the CLI parser."""
    parser = create_base_parser("xstxt", __doc__, __usages__)

    parser.add_argument(
        "-pt",
        "--title-css-path",
        default="title",
        dest="title_css_path",
        help="set css path of chapter title (default: '%(default)s')",
        type=str,
        metavar="CSS_PATH",
    )
    parser.add_argument(
        "-pb",
        "--body-css-path",
        default="body",
        dest="body_css_path",
        help="set css path of chapter body (default: '%(default)s')",
        type=str,
        metavar="CSS_PATH",
    )
    parser.add_argument(
        "-la",
        "--language",
        dest="language",
        default="zh",
        help="language of the ebook (default: '%(default)s')",
        metavar="LANGUAGE",
    )
    parser.add_argument(
        "-ps",
        "--paragraph-separator",
        dest="paragraph_separator",
        type=lambda value: value.encode("utf-8").decode("unicode_escape"),
        default="\n\n",
        help="set paragraph separator (default: %(default)r)",
        metavar="SEPARATOR",
    )
    parser.add_argument(
        "-rh",
        "--html-replace",
        default=[],
        action="append",
        dest="html_replace",
        nargs=2,
        help="set regex to replace word or pharase in html file",
        type=str,
        metavar="REGEX",
    )
    parser.add_argument(
        "-rt",
        "--txt-replace",
        default=[],
        action="append",
        dest="txt_replace",
        nargs=2,
        help="set regex to replace word or pharase in txt file",
        type=str,
        metavar="REGEX",
    )
    parser.add_argument(
        "-bt",
        "--book-title",
        default="不详",
        dest="book_title",
        help="set title of the novel (default: '%(default)s')",
        type=str,
        metavar="TITLE",
    )
    parser.add_argument(
        "-ba",
        "--book-author",
        default="不详",
        dest="book_author",
        help="set author of the novel (default: '%(default)s')",
        type=str,
        metavar="AUTHOR",
    )
    parser.add_argument(
        "-ic",
        "--indent-chars",
        default="",
        dest="indent_chars",
        help=(
            "set indent characters for a paragraph (default: '%(default)s')"
        ),
        type=str,
        metavar="INDENT_CHARS",
    )
    parser.add_argument(
        "-fw",
        "--fullwidth",
        default=False,
        action="store_true",
        dest="fullwidth",
        help=(
            "convert ASCII character to from halfwidth to fullwidth "
            "(default: '%(default)s')"
        ),
    )
    parser.add_argument(
        "-oi",
        "--output-individual-file",
        default=False,
        action="store_true",
        dest="output_individual_file",
        help="convert each html file into own txt file",
    )
    parser.add_argument(
        "-ow",
        "--overwrite",
        default=False,
        action="store_true",
        dest="overwrite",
        help="overwrite output file",
    )

    parser.add_argument(
        "-i",
        "--input",
        default=["./*.html"],
        action="append",
        dest="input",
        help=(
            "set glob pattern of html files to process "
            "(default: '%(default)s')"
        ),
        type=str,
        metavar="GLOB_PATTERN",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        default=[],
        action="append",
        dest="exclude",
        help=(
            "set glob pattern of html files to exclude "
            "(default: '%(default)s')"
        ),
        type=str,
        metavar="GLOB_PATTERN",
    )
    parser.add_argument(
        "-l",
        "--limit",
        default=3,
        dest="limit",
        help="set number of html files to process (default: '%(default)s')",
        type=int,
        metavar="TOTAL_FILES",
    )
    parser.add_argument(
        "-w",
        "--width",
        default=0,
        dest="width",
        help="set the line width for wrapping "
        "(default: %(default)s, 0 to disable)",
        type=int,
        metavar="WIDTH",
    )
    parser.add_argument(
        "-o",
        "--output",
        default="book.txt",
        dest="output",
        help="set output txt file name (default: '%(default)s')",
        type=str,
        metavar="FILENAME",
    )
    parser.add_argument(
        "-od",
        "--output-dir",
        default="output",
        dest="output_dir",
        help="set output directory (default: '%(default)s')",
        type=str,
        metavar="OUTPUT_DIR",
    )

    parser.add_argument(
        "-y",
        "--yes",
        default=False,
        action="store_true",
        dest="yes",
        help="yes to prompt",
    )

    parser.add_argument(
        "-p",
        "--purge",
        default=False,
        action="store_true",
        dest="purge",
        help=(
            "remove extracted files specified by --output-folder option "
            "(default: '%(default)s')"
        ),
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "-g",
        "--generate-config-file",
        nargs="?",
        default=False,
        const=DEFAULT_CONFIG_FILE,
        dest="generate_config",
        help="generate config file from options (default: '%(const)s')",
        type=str,
        metavar="FILENAME",
    )

    group.add_argument(
        "-c",
        "--config-file",
        nargs="?",
        default=False,
        const=DEFAULT_CONFIG_FILE,
        dest="config",
        help="load config from file (default: '%(const)s')",
        type=str,
        metavar="FILENAME",
    )

    parser.add_argument(
        "-m",
        "--monitor",
        default=False,
        action="store_true",
        dest="monitor",
        help="monitor config file changes and re-run when needed",
    )

    return parser


def main(args: Sequence[str] | None = None) -> None:
    """Run the main program flow."""
    args = args or sys.argv[1:]
    _logger.debug(args)

    try:
        config = _load_config(args)
        _logger.debug(config)

        asyncio.run(generate_book(config), debug=config.debug)
        if config.monitor:
            _run_monitor(config)
    except Exception as error:
        _logger.error(
            "error: %s",
            getattr(error, "message", str(error)),
            exc_info=("-d" in args or "--debug" in args),
        )
        raise SystemExit(1) from None


def cli() -> None:
    """Set the main entrypoint of the console app."""
    main(sys.argv[1:])


def _compile_regex_replacements(
    regexs: list[tuple[str, str]],
) -> list[tuple[re.Pattern[str], str]]:
    """Compile a list of regex search/replace pairs."""
    compiled_regexs = []
    for search, replace in regexs:
        _logger.debug("search: %s -> replace: %s", repr(search), repr(replace))
        try:
            compiled_regexs.append(
                (re.compile(rf"{search}", re.MULTILINE), rf"{replace}"),
            )  # pylint: disable=no-member
        except re.error as error:
            _logger.error("Invalid regex: %s", error)
            raise
    return compiled_regexs


def _run_monitor(config: argparse.Namespace) -> None:
    """Monitor the config file changes in the background.

    Args:
        config(dict): Config from file.
    """
    observer = Observer()
    try:
        _logger.info("Running in monitor mode")
        event_handler = ConfigFileHandler(config)
        observer.schedule(event_handler, path=".")
        observer.start()

        while observer.is_alive():
            observer.join(1)
    finally:
        observer.stop()
        observer.join()


def _load_config(args: Sequence[str] | None = None) -> argparse.Namespace:
    """Load configuration from file and command line arguments."""
    parser = build_parser()
    parsed_args = parser.parse_args(args)

    setup_logging(parsed_args)

    config_from_file = load_or_create_config(parsed_args, "xstxt")
    parser.set_defaults(**config_from_file)
    config = parser.parse_args()

    # Compile regexes once
    config.compiled_html_replace = _compile_regex_replacements(
        config.html_replace,
    )
    config.compiled_txt_replace = _compile_regex_replacements(
        config.txt_replace,
    )

    config.gettext_func = _load_translation(config)
    return config


def _load_translation(config: argparse.Namespace) -> Callable[[str], str]:
    """Load translation for the given language."""
    localedir = Path(__file__).parent / "locales"
    _logger.debug("locale directory: %s", localedir)

    translation = gettext.translation(
        "xstxt",
        localedir=str(localedir),
        languages=[config.language],
    )
    return translation.gettext


if __name__ == "__main__":  # pragma: no cover
    cli()
